
#define _itmtouch_C_

#include <misc.h>
#include <xf86.h>
#define NEED_XF86_TYPES
#include <xf86_ansic.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <xisb.h>
#include <exevents.h>
#include "itmtouch.h"

/*===========================================================================
        Constants
 ===========================================================================*/

#define ITM_CALIBRATION_STATUS_1 0xC1
#define ITM_CALIBRATION_STATUS_2 0xC2
#define ITM_CALIBRATION_ERROR    0xC0

/*===========================================================================
        Structures
 ===========================================================================*/

InputDriverRec ITMTOUCH = {
     1,
     "itmtouch",
     NULL,
     ItmTouchPreInit,
     /*ItmTouchUnInit*/ NULL,
     NULL,
     0
};

#ifdef XFree86LOADER

static XF86ModuleVersionInfo VersionRec =
{
	"itmtouch",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	1, 0, 0,
	ABI_CLASS_XINPUT,
	ABI_XINPUT_VERSION,
	MOD_CLASS_XINPUT,
	{0, 0, 0, 0}				/* signature, to be patched into the file by
								 * a tool */
};

static const char *reqSymbols[] = {
	"AddEnabledDevice",
	"ErrorF",
	"InitButtonClassDeviceStruct",
	"InitProximityClassDeviceStruct",
	"InitValuatorAxisStruct",
	"InitValuatorClassDeviceStruct",
	"RemoveEnabledDevice",
	"Xcalloc",
	"Xfree",
	"XisbBlockDuration",
	"XisbFree",
	"XisbNew",
	"XisbRead",
	"XisbTrace",
	"XisbWrite",
	"screenInfo",
	"xf86AddInputDriver",
	"xf86AllocateInput",
	"xf86CloseSerial",
	"xf86CollectInputOptions",
	"xf86ErrorF",
	"xf86ErrorFVerb",
	"xf86FindOptionValue",
	"xf86FlushInput",
	"xf86GetMotionEvents",
	"xf86GetVerbosity",
	"xf86LoaderReqSymLists",
	"xf86MotionHistoryAllocate",
	"xf86Msg",
	"xf86NameCmp",
	"xf86OpenSerial",
	"xf86OptionListCreate",
	"xf86OptionListFree",
	"xf86OptionListReport",
	"xf86PostButtonEvent",
	"xf86PostMotionEvent",
	"xf86PostProximityEvent",
	"xf86ProcessCommonOptions",
	"xf86ScaleAxis",
	"xf86SetIntOption",
	"xf86SetSerial",
	"xf86SetStrOption",
	"xf86XInputSetScreen",
	"xf86XInputSetSendCoreEvents",
	"xf86memset",
	"xf86sscanf",
	"xf86strcmp",
	"xf86strlen",
	"xf86strncmp",
	"xf86strncpy",
        NULL
};

static pointer
SetupProc(	pointer module,
		pointer options,
		int *errmaj,
		int *errmin )
{
        xf86LoaderReqSymLists(reqSymbols, NULL);
	xf86AddInputDriver(&ITMTOUCH, module, 0);
	return (pointer) 1;
}

XF86ModuleData itmtouchModuleData = {&VersionRec, &SetupProc, NULL };

#endif /* XFree86LOADER */

static const char *default_options[] =
{
	"BaudRate", "9600",
	"StopBits", "1",
	"DataBits", "8",
	"Parity",   "None",
        "SwapXY",   "false",
};


/*===========================================================================
 	Function Definitions
 ===========================================================================*/


static InputInfoPtr
ItmTouchPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
	LocalDevicePtr local;
	ItmTouchPrivatePtr priv;

	char *s;

	priv = xcalloc (1, sizeof (ItmTouchPrivateRec));
	if (!priv)
		return NULL;

	local = xf86AllocateInput(drv, 0);
	if (!local) {
		xfree(priv);
		return NULL;
	}

	local->type_name = XI_TOUCHSCREEN;
	local->device_control = DeviceControl;
	local->read_input = ReadInput;
	local->control_proc = ControlProc;
	local->close_proc = CloseProc;
	local->switch_mode = SwitchMode;
	local->conversion_proc = ConvertProc;
	local->dev = NULL;
	local->private = priv;
	local->private_flags = 0;
	local->flags = XI86_POINTER_CAPABLE; /*TODO ? | XI86_SEND_DRAG_EVENTS;*/
	local->conf_idev = dev;

	xf86CollectInputOptions(local, default_options, NULL);

	xf86OptionListReport(local->options);

	local->fd = xf86OpenSerial (local->options);
	if (local->fd == -1)
	{
		ErrorF ("ItmTouch driver unable to open device\n");
		goto SetupProc_fail;
	}
	xf86ErrorFVerb( 6, "tty port opened successfully\n" );

	priv->min_x = xf86SetIntOption( local->options, "MinX", 0 );
	priv->max_x = xf86SetIntOption( local->options, "MaxX", 4096 );
	priv->min_y = xf86SetIntOption( local->options, "MinY", 0 );
	priv->max_y = xf86SetIntOption( local->options, "MaxY", 4096 );
	      
        priv->screen_num = xf86SetIntOption( local->options, "ScreenNumber", 0 );
	priv->button_number = xf86SetIntOption( local->options, "ButtonNumber", 1 );

	s = xf86FindOptionValue (local->options, "SwapXY");
	if ((s) && (xf86NameCmp (s, "true") == 0))
		priv->swap_axis = TRUE;
	else
		priv->swap_axis = FALSE;
      
	priv->buffer = XisbNew (local->fd, 200);
	priv->synchronized = FALSE;
	priv->reading_a_packet = FALSE;
        priv->partial_packet_i = 0;
   
        priv->button_down = FALSE;

	DBG (9, XisbTrace (priv->buffer, 1));
	
        if (QueryHardware(local) != Success)
	{
		ErrorF ("Unable to query/initialize ItmTouch hardware.\n");
		goto SetupProc_fail;
	}

	local->history_size = xf86SetIntOption( local->options, "HistorySize", 0 );

	/* prepare to process touch packets */
        priv->synchronized = FALSE;
	priv->reading_a_packet = FALSE;
        priv->partial_packet_i = 0;

	/* this results in an xstrdup that must be freed later */
	local->name = xf86SetStrOption( local->options, "DeviceName", "ItmTouch TouchScreen" );
	xf86ProcessCommonOptions(local, local->options);
	local->flags |= XI86_CONFIGURED;

	if (local->fd != -1)
	{ 
		RemoveEnabledDevice (local->fd);
		if (priv->buffer)
		{
			XisbFree(priv->buffer);
			priv->buffer = NULL;
		}
		xf86CloseSerial(local->fd);
	}
	RemoveEnabledDevice (local->fd);
	local->fd = -1;
	return (local);

  SetupProc_fail:
	if ((local) && (local->fd))
		xf86CloseSerial (local->fd);
	if ((local) && (local->name))
		xfree (local->name);

	if ((priv) && (priv->buffer))
		XisbFree (priv->buffer);
	if (priv)
		xfree (priv);
	return (local);
}

static Bool
DeviceControl (DeviceIntPtr dev,
			   int mode)
{
	Bool	RetValue;

	switch (mode)
	{
	case DEVICE_INIT:
		DeviceInit (dev);
		RetValue = Success;
		break;
	case DEVICE_ON:
		RetValue = DeviceOn( dev );
		break;
	case DEVICE_OFF:
	case DEVICE_CLOSE:
		RetValue = DeviceOff( dev );
		break;
	default:
		RetValue = BadValue;
	}

	return( RetValue );
}

static Bool
DeviceOn (DeviceIntPtr dev)
{
	LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
	ItmTouchPrivatePtr priv = (ItmTouchPrivatePtr) (local->private);

	local->fd = xf86OpenSerial(local->options);
	if (local->fd == -1)
	{
		xf86Msg(X_WARNING, "%s: cannot open input device\n", local->name);
		return (!Success);
	}

	priv->buffer = XisbNew(local->fd, 64);
	if (!priv->buffer) 
		{
			xf86CloseSerial(local->fd);
			local->fd = -1;
			return (!Success);
		}

	xf86FlushInput(local->fd);
	AddEnabledDevice (local->fd);
	dev->public.on = TRUE;
	return (Success);
}

static Bool
DeviceOff (DeviceIntPtr dev)
{
	LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
	ItmTouchPrivatePtr priv = (ItmTouchPrivatePtr) (local->private);

	if (local->fd != -1)
        { 
	RemoveEnabledDevice (local->fd);
		if (priv->buffer)
		{
			XisbFree(priv->buffer);
			priv->buffer = NULL;
                }
		xf86CloseSerial(local->fd);
	}

	RemoveEnabledDevice (local->fd);
	dev->public.on = FALSE;
	return (Success);
}

static Bool
DeviceInit (DeviceIntPtr dev)
{
	LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
	ItmTouchPrivatePtr priv = (ItmTouchPrivatePtr) (local->private);
	unsigned char map[] =
	{0, 1};

	/* 
	 * these have to be here instead of in the SetupProc, because when the
	 * SetupProc is run at server startup, screenInfo is not setup yet
	 */
	priv->screen_width = screenInfo.screens[priv->screen_num]->width;
	priv->screen_height = screenInfo.screens[priv->screen_num]->height;

	/* 
	 * Device reports button press for 1 button.
	 */
	if (InitButtonClassDeviceStruct (dev, priv->button_number, map) == FALSE)
	{
		ErrorF ("Unable to allocate ItmTouch touchscreen ButtonClassDeviceStruct\n");
		return !Success;
	}

	/* 
	 * Device reports motions on 2 axes in absolute coordinates.
	 */
	if (InitValuatorClassDeviceStruct (dev, 2, xf86GetMotionEvents,
									local->history_size, Absolute) == FALSE)
	{
		ErrorF ("Unable to allocate ItmTouch touchscreen ValuatorClassDeviceStruct\n");
		return !Success;
	}
	else
	{       
		InitValuatorAxisStruct (dev, 0, priv->min_x, priv->max_x,
								9500,
								0 /* min_res */ ,
								9500 /* max_res */ );
		InitValuatorAxisStruct (dev, 1, priv->min_y, priv->max_y,
								10500,
								0 /* min_res */ ,
								10500 /* max_res */ );
	}
        
	/* 
	 * Allocate the motion events buffer.
	 */
	xf86MotionHistoryAllocate (local);
	return (Success);
}

 
static void
ReadInput (LocalDevicePtr local)
{
	int x, y;
	int pressure;
        Bool lift_off;
   
	ItmTouchPrivatePtr priv = (ItmTouchPrivatePtr) (local->private);

	/* 
	 * set blocking to -1 on the first call because we know there is data to
	 * read. Xisb automatically clears it after one successful read so that
	 * succeeding reads are preceeded buy a select with a 0 timeout to prevent
	 * read from blocking indefinately.
	 */
	XisbBlockDuration (priv->buffer, -1);
	while (ItmTouchGetPacket (priv) == Success)
	{

	   lift_off = ( priv->packet[0] & 0x40) ? TRUE : FALSE; 
	   
	   if ( lift_off == FALSE )
	     {	
		
		y = (((unsigned short)priv->packet[0]) << 7 ) & 0x0F80 ;
		x = (((unsigned short)priv->packet[1]) << 7 ) & 0x0F80 ;
		
		y |= ((unsigned short)priv->packet[2]) & 0x007F; 
		x |= ((unsigned short)priv->packet[3]) & 0x007F;
		
		pressure = ( priv->packet[4] & 0x35 ) | (( priv->packet[1] << 1 ) & 0xC0);
		
		if ( priv->swap_axis)
		  {
		     int tmp;
		     tmp = x;
		     x = y;
		     y = tmp;
		  }		
		
		x = xf86ScaleAxis (x, 0, priv->screen_width, priv->min_x,
				   priv->max_x);
		y = xf86ScaleAxis (y, 0, priv->screen_height, priv->min_y,
				   priv->max_y);
	     

		xf86XInputSetScreen (local, priv->screen_num, x, y);
	    
		/* 
		 * Send events.
		 *
		 * We *must* generate a motion before a button change if pointer
		 * location has changed as DIX assumes this. This is why we always
		 * emit a motion, regardless of the kind of packet processed.
		 */
		
		xf86PostMotionEvent (local->dev, TRUE, 0, 2, x, y);

		/* 
		 * Emit a button press or release.
		 */
		if (priv->button_down == FALSE)

		{
			xf86PostButtonEvent (local->dev, TRUE,
					     priv->button_number, 1, 0, 2, x, y);
			priv->button_down = TRUE;
		}

		/*
		 * We need to save x and y values because the lift off packets 
		 * don't contains them.
		 */
		
		priv->last_x = x;
		priv->last_y = y;
		
		/*xf86ErrorFVerb( 3, "TouchScreen: %5d,%5d (%5d) %s\n",
						x, y, pressure,
						(pressure > 1) ? "Press" : "Release" );*/
	     }
	   else /* LIFT_OFF packet */
	     {
		/* 
		 * the lift off should always come after the release
		 */

		if (priv->button_down == TRUE)
		  {
		     xf86PostButtonEvent (local->dev, TRUE,
					  priv->button_number, 0, 0, 2, priv->last_x, priv->last_y);
		     priv->button_down = FALSE;
		  }	
	     }	   
	}
}

static int
ControlProc (LocalDevicePtr local, xDeviceCtl * control)
{
	xDeviceTSCalibrationCtl *c = (xDeviceTSCalibrationCtl *) control;
	ItmTouchPrivatePtr priv = (ItmTouchPrivatePtr) (local->private);

	priv->min_x = c->min_x;
	priv->max_x = c->max_x;
	priv->min_y = c->min_y;
	priv->max_y = c->max_y;
	return (Success);
}

static void
CloseProc (LocalDevicePtr local)
{
}

static int
SwitchMode (ClientPtr client, DeviceIntPtr dev, int mode)
{
	LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
	
	if ((mode == SendCoreEvents) || (mode == DontSendCoreEvents))
	{
		xf86XInputSetSendCoreEvents (local, (mode == SendCoreEvents));
		return (Success);
	}
	else
		return (!Success);
}

static Bool
ConvertProc (LocalDevicePtr local,
			 int first,
			 int num,
			 int v0,
			 int v1,
			 int v2,
			 int v3,
			 int v4,
			 int v5,
			 int *x,
			 int *y)
{
/*	ItmTouchPrivatePtr priv = (ItmTouchPrivatePtr) (local->private);

	if (priv->reporting_mode == TS_Raw)
	{
		*x = xf86ScaleAxis (v0, 0, priv->screen_width, priv->min_x,
							priv->max_x);
		*y = xf86ScaleAxis (v1, 0, priv->screen_height, priv->min_y,
							priv->max_y);
	}
	else
*/	{
		*x = v0;
		*y = v1;
	}
	return (TRUE);
}

static Bool
xf86ItmTouchSendCommand (unsigned char *type, ItmTouchPrivatePtr priv)
{
	return (FALSE);
}

static Bool
QueryHardware (LocalDevicePtr local)
{
        /*
	 * TODO: use serial PNP protocol to ensure the connected device
	 * is a itm_touch controler
	 */
        return (Success);
}


static Bool
xf86ItmTouchSendPacket (unsigned char *type, int len, ItmTouchPrivatePtr priv)
{
        /* Not implemented yet */
	return Success;
}

/* 
 ***************************************************************************
 *   0 ACK
 *  -1 NACK
 *  -2 timeout
 *  -3 wrong packet type
 ***************************************************************************
 */

static int
xf86ItmTouchWaitReply (unsigned char *type, ItmTouchPrivatePtr priv)
{
    /* Not implemented yet */  
    return (!Success);
}

static Bool
ItmTouchGetPacket (ItmTouchPrivatePtr priv)
{
   int c;
   
   while ((c = XisbRead (priv->buffer)) >= 0)
     {
	
	if ( priv->synchronized == TRUE)
	  {
	 
	     priv->partial_packet[priv->partial_packet_i++] = (unsigned char)c;
	     
	     if ( c & 0x80 ) /* if the most significant bit is on we have reached the last byte of a packet */
	       {
		  int b;
		  
		  memcpy( priv->packet, priv->last_packet, ITM_FULL_PACKET_SIZE);
		  
		  for( b = 0 ; b < priv->partial_packet_i ; b++) 
		    priv->packet[b+(ITM_FULL_PACKET_SIZE-priv->partial_packet_i)] = priv->partial_packet[b];
		  
		  priv->partial_packet_i = 0;
		       
		  memcpy( priv->last_packet, priv->packet, ITM_FULL_PACKET_SIZE);
	       
		  return (Success);

	       }
	     
	     if ( priv->partial_packet_i >= ITM_FULL_PACKET_SIZE )
	       {
		  priv->synchronized = FALSE;
		  priv->reading_a_packet = FALSE;
		  xf86ErrorFVerb( 8, "Too large packet for being sent by an ItmTouch controler !\n");
	       }
	  }
	else /* Not synchronized */
	  {
	     /* To be synchronized we need to read a full packet */
	     
	     if ( priv->reading_a_packet == TRUE)
	       {
		  priv->partial_packet[priv->partial_packet_i++] = (unsigned char)c;
		  
		  if ( ( c & 0x80 ) && ( priv->partial_packet_i == ITM_FULL_PACKET_SIZE ) )
		    {
		       /* Yes, we get it ! */
		       xf86ErrorFVerb( 8, "ItmTouch SYCHRONIZED !\n");
		       priv->partial_packet_i = 0;
		       priv->synchronized = 1;
		       /* copy to last packet */
		       memcpy( priv->last_packet, priv->partial_packet, ITM_FULL_PACKET_SIZE);
		    }
		  else if ( c & 0x80 )
		    {
		       /* End of packet but unfortunately this one is not complete */
		       priv->partial_packet_i = 0;
		    }
		  else if ( priv->partial_packet_i >= ITM_FULL_PACKET_SIZE)
		    {
		       xf86ErrorFVerb( 8, "Too large packet for being sent by an ItmTouch controler !\n");
		       priv->reading_a_packet = FALSE;
		    }   
	       }
	     else /* We don't know were we are, let's waiting for the end of a packet */
	       {
		  if ( c & 0x80 ) /* the end of a packet */
		    {
		       priv->reading_a_packet = TRUE;
		       priv->partial_packet_i = 0;
		    }
		  else
		    {
		       /* Do nothing */
		    }  
	       }
	  }
     }
     	
   return (!Success);
}

/* 
 ***************************************************************************
 *
 *  xf86ItmTouchPrintIdent --
 *  Print type of touchscreen and features on controller board.
 *
 ***************************************************************************
 */
static void
xf86ItmTouchPrintIdent (unsigned char *packet)
{
   /* Not implemented yet */
}

/* 
 ***************************************************************************
 *
 *  xf86ItmTouchPrintHwStatus --
 *  Print status of hardware. That is if the controller report errors,
 *  decode and display them.
 *
 ***************************************************************************
 */
static Bool
xf86ItmTouchPrintHwStatus (unsigned char *packet)
{
       /* Not implemented yet */	
       return (Success);
}
